#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class SynchObj {
  sem_t m;
  sem_t s;
  int nblocked;
  int nthreads;
public:
  SynchObj(int n);

  void synch();
};

SynchObj::SynchObj(int n)
{
  nthreads = n;
  nblocked = 0;
  sem_init(&s, 0, 0);
  sem_init(&m, 0, 1);
}

void SynchObj::synch()
{
  sem_wait(&m);

  nblocked++;
  if (nblocked == nthreads) {
    nblocked = 0;
    for (int i=0; i<nthreads-1; i++) 
	sem_post(&s);
    sem_post(&m);  
  }
  else {
    sem_post(&m);
    sem_wait(&s);
  }
}

void *thread_body(void *arg)
{
  SynchObj *s = (SynchObj *)arg;

  sleep(rand() % 4);

  printf("Before the synchronization point...\n");

  s->synch();

  printf("After the synchronization point!\n");
}


int main(int argc, char *argv[])
{
  pthread_t *array;
  int i;

  if (argc < 2) {
    printf("Usage: %s <nthreads>\n", argv[0]);
    exit(0);
  }
  int n = atoi(argv[1]);
  printf("Example with %d threads\n", n);
  
  array = new pthread_t[n];

  SynchObj my_synch(n);

  for(i=0; i<n; i++) 
    pthread_create(&array[i], 0, thread_body, (void *)(&my_synch));
  
  for (i=0; i<n; i++) 
    pthread_join(array[i], 0);

}
