#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class CircularArray {
    pthread_mutex_t m;
	
    pthread_cond_t prod;
    pthread_cond_t cons;
    pthread_cond_t manag;

    int nelem;
    int maxelem;
    int fermo;

    int head;
    int tail;

    int *array;

public:
    CircularArray(int size);

    void insert(int elem);
    void extract(int &elem);

    void start();
    void stop();
};

CircularArray::CircularArray(int size)
{
    array = new int[size];
    maxelem = size;
    head = tail = 0;

    fermo = 0; nelem = 0;

    pthread_mutex_init(&m, 0);

    pthread_cond_init(&prod, 0);
    pthread_cond_init(&cons, 0);
    pthread_cond_init(&manag, 0);
}

void CircularArray::insert(int elem)
{
    pthread_mutex_lock(&m);

    while (nelem == maxelem || fermo) 
        pthread_cond_wait(&prod, &m);

    array[head] = elem;
    head = (head + 1) % maxelem;

    nelem++;

    pthread_cond_signal(&cons);

    pthread_mutex_unlock(&m);
}

void CircularArray::extract(int &elem)
{
    pthread_mutex_lock(&m);

    while (nelem == 0) 
        pthread_cond_wait(&cons, &m);

    elem = array[tail];
    tail = (tail + 1) % maxelem;

    nelem --;

    if (fermo == 1 && nelem == 0) 
        pthread_cond_signal(&manag);
    else if (fermo == 0) 
        pthread_cond_signal(&prod);
		
    pthread_mutex_unlock(&m);
}

void CircularArray::stop()
{
	pthread_mutex_lock(&m);
	fermo = 1;
	pthread_mutex_unlock(&m);	
}

void CircularArray::start()
{
	pthread_mutex_lock(&m);

	if (fermo == 1 && nelem > 0) 
		pthread_cond_wait(&manag, &m);

	fermo = 0;

	pthread_cond_broadcast(&prod);

	pthread_mutex_unlock(&m);
}

CircularArray ca(5);

void *producer(void *arg)
{
    int counter = 0;
    while (1) {
        printf("PROD: inserting!\n");
        ca.insert(counter);
        counter++;
        sleep(2);
    } 
    return 0;
}

void *consumer(void *arg)
{
    int data;
    while (1) {
        printf("CONS: extracting!\n");
        ca.extract(data);
        sleep(2);
    }
    return 0;
}


#define NCONS   2
#define NPROD   4

int main()
{
  int i;
  pthread_t prod_tid[NPROD];
  pthread_t cons_tid[NCONS];

  printf("Press 's' for fermo, 'b' for start, 'q' to exit\n");

  for(i=0; i<NPROD; i++)
    pthread_create(&prod_tid[i], 0, producer, 0);

  for(i=0; i<NCONS; i++)
    pthread_create(&cons_tid[i], 0, consumer, 0);

  while(1) {
    char c;

    c = getchar();

    if (c == 's') {
      printf("fermo command! \n");
      ca.stop();
    }
    else if (c == 'b') {
      printf("Start command! \n");
      ca.start();
      printf("Start command completed! \n");
    }
    else if (c == 'q') exit(-1);
  }
}














