#include <stdio.h>
#include <semaphore.h>
#include <pthread.h>
#include <stdlib.h>
#include <unistd.h>

#define FROM_RIGHT    1
#define FROM_LEFT     0

#define MAX_CAR       3

class Road {
  sem_t m;
  sem_t sdir[2];
  sem_t p, c;
  int nblocked[2];
public:
  Road();
  
  int wait_for_car(int dir);
  void wait_for_passage();

  void request_passage(int dir);
  void go_out();
}; 


Road::Road()
{
  sem_init(&m, 0, 1);
  sem_init(&p, 0, 0);
  sem_init(&c, 0, 0);
  for (int i=0; i<2; i++) {
    nblocked[i] = 0; 
    sem_init(&sdir[i], 0, 0);
  }
}

int Road::wait_for_car(int dir)
{
  sem_wait(&p);
  sem_wait(&m);
  if (nblocked[dir] == 0) dir = (dir + 1) % 2;

  nblocked[dir]--;
  sem_post(&m);
  sem_post(&sdir[dir]);
  return dir;
}

void Road::request_passage(int dir)
{
  sem_post(&p);
  sem_wait(&m);
  nblocked[dir]++;
  sem_post(&m);
  sem_wait(&sdir[dir]);
} 


void Road::wait_for_passage()
{
  sem_wait(&c);
}

void Road::go_out()
{
  sem_post(&c);
}


Road r;

void *policeman(void *)
{
    int curr_dir = FROM_LEFT;
    int dir = curr_dir; 
    int count = 0;

    while(1) {
        dir = r.wait_for_car(curr_dir);

        if (dir == curr_dir) {
            count++;
            printf("Car #%d in direction %s\n", count, ((dir == FROM_RIGHT) ? "right" : "left"));
            if (count == MAX_CAR) {
                count = 0;
                curr_dir = (curr_dir + 1) % 2;
                printf("Changing direction!\n");
            }
        }
        else {
            printf("No more cars in current direction, changing direction!\n");
            count = 1;
            curr_dir = dir;
        }
        r.wait_for_passage();
    }
    return 0;
}

void *car(void *arg)
{
    int dir = *((int*)arg);

    sleep(rand() % 3);

    printf("Car coming from %s\n", ((dir == FROM_RIGHT) ? "right" : "left"));
    r.request_passage(dir);
    printf("Car coming from %s is passing\n", ((dir == FROM_RIGHT) ? "right" : "left"));

    sleep(3);

    r.go_out();
    return 0;
}


int main()
{
    pthread_t tid[16];
    int index[16];
    pthread_t pol;
    int i;
 
    pthread_create(&pol, 0, policeman, 0);

    for (i=0; i<16; i++) {
        index[i] = i%2;
        pthread_create(&tid[i], 0, car, &index[i]);
    } 

    for (i=0; i<16; i++) 
        pthread_join(tid[i], 0);

}



