#include <pthread.h>
#include <semaphore.h>
#include <iostream>
#include <stdlib.h>

using namespace std;

#define FROM_LEFT   0
#define FROM_RIGHT  1

#define MAX         3

class Narrow {
public:
    Narrow();
    void request_passage(int dir);
    void go_out();
private:

    int blocked_right;
    int blocked_left;
    sem_t bright;
    sem_t bleft;
    sem_t mutex;

    int direction;
    int passed;
    int passing;
    int changing;
};

Narrow::Narrow()
{
    blocked_right = 0;
    blocked_left = 0;
    changing = 0;
    sem_init(&bright, 0, 0);
    sem_init(&bleft, 0, 0);
    sem_init(&mutex, 0, 1);

    passed = 0;
    passing = 0;
    direction = FROM_LEFT;
}

void Narrow::request_passage(int dir)
{
    sem_wait(&mutex);

    while (dir != direction || changing == 1) {
        if (dir == FROM_LEFT && passing == 0 && 
            blocked_right == 0 && changing == 0) {
            // change direction
            direction = FROM_LEFT;
            cout << "Direction changed: " << direction << endl;
            cout << "Passed: " << passed << endl; 
            passed = 0;
        } 
        else if (dir == FROM_RIGHT && passing == 0 && 
                 blocked_left == 0 && changing == 0) {
            // change direction
            direction = FROM_RIGHT;
            cout << "Direction changed: " << direction << endl;
            cout << "Passed: " << passed << endl; 
            passed = 0;
        }
        else {
            // block
            if (dir == FROM_LEFT) {
                blocked_left++;
                sem_post(&mutex);
                sem_wait(&bleft);
            }
            else if (dir == FROM_RIGHT) {
                blocked_right++;
                sem_post(&mutex);
                sem_wait(&bright);
            }
            sem_wait(&mutex);
        }
    }

    passing++;
    passed++;
    
    if (passed >= MAX && dir == FROM_LEFT && blocked_right > 0) {
        changing = 1;
        direction = FROM_RIGHT;
        cout << "Changing direction: " << direction << endl;
        cout << "Passed: " << passed << endl;
        passed = 0;
    }
    else if (passed >= MAX && dir == FROM_RIGHT && blocked_left >0) {
        changing = 1;
        direction = FROM_LEFT;
        cout << "Changing direction: " << direction << endl;
        cout << "Passed: " << passed << endl;
        passed = 0;
    }

    sem_post(&mutex);
}


void Narrow::go_out()
{
    sem_wait(&mutex);
    int i;

    passing--;

    if (passing == 0) {  
        if (changing == 0 && direction == FROM_LEFT &&
            blocked_left == 0 && blocked_right > 0) {
            direction = FROM_RIGHT;
            changing = 1;
            cout << "Changing direction: " << direction << endl;
            cout << "Passed: " << passed << endl;
            passed = 0;
        }
        else if (changing == 0 && direction == FROM_RIGHT &&
            blocked_right == 0 && blocked_left > 0) {
            direction = FROM_LEFT;
            changing = 1;
            cout << "Changing direction: " << direction << endl;
            cout << "Passed: " << passed << endl;
            passed = 0;
        }

        if (changing == 1) {
            if (direction == FROM_RIGHT) {
                for (i=0; i<blocked_right; i++) {
                    sem_post(&bright);
                    if (i>=MAX) break;
                }
                blocked_right -= i;
            }
            else if (direction == FROM_LEFT) {
                for (i=0; i<blocked_left; i++) {
                    sem_post(&bleft);
                    if (i>=MAX) break;
                }
                blocked_left -= i;
            }
            changing = 0;
            cout << "Direction changed: " << direction << endl;
        }
    }

    sem_post(&mutex);
}


Narrow street;

void pass()
{
    sleep(2);
}

#define NCARS 20

void *thread(void *arg)
{
    int index = *((int *) arg);

    int dir = index % 2;

    sleep(rand()%3);

    cout << "Car " << index << " requesting passage" << endl;
    
    street.request_passage(dir);

    cout << "Car " << index << " entering the street" << endl;

    pass();

    cout << "Car " << index << " exiting the street" << endl;

    street.go_out();
    
    return 0;
}


int main()
{
    pthread_t tid[NCARS];
    int index[NCARS];
    int i;

    cout << "Creating threads" << endl;

    for (int i=0; i<NCARS; i++) {
        index[i] = i;

        pthread_create(&tid[i], 0, thread, &index[i]);
    }

    cout << "Waiting for termination" << endl;

    for (int i=0; i<NCARS; i++) 
        pthread_join(tid[i], 0);

}
